from unittest.mock import MagicMock

import pytest

from .config import HtmlDisplayConfiguration
from .publisher import Publisher


@pytest.fixture
def publisher():
    return Publisher(
        actor=MagicMock(actor="http://actor.test/", name="name"),
        config=HtmlDisplayConfiguration(),
        obj={},
    )


def test_id_stays_idempotent(publisher):
    stored_object = publisher.object_for_store

    new_publisher = Publisher(
        actor=MagicMock(actor="http://actor.test/", name="name"),
        config=HtmlDisplayConfiguration(),
        obj=stored_object,
    )

    new_id = new_publisher.object_for_store.get("id")

    assert new_id == stored_object.get("id")
    assert publisher.uuid == new_publisher.uuid


def test_create_html_link(publisher):
    result = publisher.create_html_link()

    assert isinstance(result, dict)


def test_object_for_remote(publisher) -> None:
    obj = publisher.object_for_remote

    assert isinstance(obj.get("url"), list)

    object_id = obj.get("id")
    assert isinstance(object_id, str)
    assert obj.get("replies") == object_id + "/replies"
    assert obj.get("shares") == object_id + "/shares"
    assert obj.get("likes") == object_id + "/likes"



import logging
import random
import string
from uuid import UUID

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from cattle_grid.extensions.examples.html_display.database import (
    PublishedObject,
    PublishingActor,
)


logger = logging.getLogger(__name__)


def new_name():
    return "".join(random.choices(string.ascii_letters, k=12))


async def publishing_actor_for_actor_id(session: AsyncSession, actor_id: str):
    actor = await session.scalar(
        select(PublishingActor).where(PublishingActor.actor == actor_id)
    )
    if actor:
        return actor

    actor = PublishingActor(actor=actor_id, name=new_name())
    session.add(actor)
    return actor


async def object_for_object_id(
    session: AsyncSession, object_id: str
) -> PublishedObject | None:
    try:
        object_uuid = UUID(object_id.split("/")[-1])
    except Exception:
        return None

    return await session.scalar(
        select(PublishedObject).where(PublishedObject.id == object_uuid)
    )



import logging
from uuid import UUID
from bovine.activitystreams import OrderedCollection
import jinja2
from fastapi.templating import Jinja2Templates

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from sqlalchemy import select

from muck_out.transform.utils import sanitize_html

from cattle_grid.activity_pub import is_valid_requester_for_obj
from cattle_grid.dependencies.fastapi import SqlSession
from cattle_grid.extensions.examples.html_display.format import format_actor_profile
from cattle_grid.extensions.examples.html_display.database import (
    ExportPermission,
    PublishedObject,
    PublishingActor,
)
from cattle_grid.extensions.examples.html_display.publisher import Publisher
from cattle_grid.extensions.examples.html_display.types import InteractionType
from cattle_grid.tools.fastapi import (
    ActivityPubHeaders,
    ActivityResponse,
    ShouldServe,
    ContentType,
)

from cattle_grid.version import __version__

from .fastapi_dependencies import (
    ActorProfile,
    PublishedObjectForUUID,
    PublishedObjectWithInteractionsForUUID,
    PublishingActorForName,
)
from .extension_declaration import extension

logger = logging.getLogger(__name__)

templates = Jinja2Templates(
    env=jinja2.Environment(auto_reload=True, loader=jinja2.PackageLoader(__name__)),
)

router = APIRouter()


@router.get("/")
async def get_index():
    """Dummy index page, only useful for debugging"""
    return "html extension"


@router.get("/object/{uuid}", response_class=ActivityResponse, tags=["activity_pub"])
async def get_object(
    obj: PublishedObjectForUUID,
    ap_headers: ActivityPubHeaders,
    session: SqlSession,
    config: extension.ConfigFastAPI,  # type:ignore
):
    """Returns the stored object"""
    if not ap_headers.x_cattle_grid_requester:
        raise HTTPException(401)

    if ap_headers.x_ap_location != obj.data.get("id"):
        raise HTTPException(404)

    if not await is_valid_requester_for_obj(
        session, ap_headers.x_cattle_grid_requester, obj.data
    ):
        raise HTTPException(401)

    actor = await session.scalar(
        select(PublishingActor).where(PublishingActor.actor == obj.actor)
    )
    if actor is None:
        raise HTTPException(404)

    publisher = Publisher(actor, config, obj=obj.data)

    return publisher.object_for_remote


@router.get(
    "/object/{uuid}/{interaction}",
    response_class=ActivityResponse,
    tags=["activity_pub"],
)
async def get_object_interaction(
    obj: PublishedObjectWithInteractionsForUUID,
    ap_headers: ActivityPubHeaders,
    session: SqlSession,
    interaction: InteractionType,
):
    """Returns the stored object"""
    if not ap_headers.x_cattle_grid_requester:
        raise HTTPException(401)

    collection_id = f"{obj.data.get('id')}/{interaction}"

    if ap_headers.x_ap_location != collection_id:
        raise HTTPException(404)

    if not await is_valid_requester_for_obj(
        session, ap_headers.x_cattle_grid_requester, obj.data
    ):
        raise HTTPException(401)

    logger.info(obj.interactions)

    return OrderedCollection(
        id=collection_id,
        items=[x.object_id for x in obj.interactions if x.interaction == interaction],
    ).build()


@router.get("/html/{actor_name}", response_class=HTMLResponse, tags=["html"])
@router.get("/html/{actor_name}/", response_class=HTMLResponse, tags=["html"])
async def get_actor_html(
    actor: PublishingActorForName,
    profile: ActorProfile,
    request: Request,
    session: SqlSession,
    should_serve: ShouldServe,
):
    """Returns the HTML representation of the actor"""

    if (
        ContentType.html not in should_serve
        and ContentType.activity_pub in should_serve
    ):
        return RedirectResponse(profile["id"])

    published_objects = await session.scalars(
        select(PublishedObject)
        .where(PublishedObject.actor == actor.actor)
        .order_by(PublishedObject.create_date.desc())
        .limit(10)
    )

    posts = [
        {
            "body": sanitize_html(x.data.get("content")),
            "date": x.data.get("published"),
            "id": str(x.id),
        }
        for x in published_objects
    ]

    return templates.TemplateResponse(
        request,
        name="index.html.j2",
        context={
            "name": actor.name,
            "profile": format_actor_profile(profile),
            "posts": posts,
            "version": __version__,
        },
    )


@router.get("/html/{actor_name}/export", tags=["html"])
async def export_objects(
    actor: PublishingActorForName, token: str, session: SqlSession
):
    """Returns the export of user data, a token needs to be provided.
    The token can be requested via the `html_display_export` method."""
    try:
        token_uuid = UUID(token)
    except ValueError:
        raise HTTPException(422, detail="Invalid uuid as token")

    permission = await session.scalar(
        select(ExportPermission)
        .where(ExportPermission.publishing_actor == actor)
        .where(ExportPermission.one_time_token == token_uuid)
    )

    if not permission:
        raise HTTPException(401)

    items = await session.scalars(
        select(PublishedObject.data).where(PublishedObject.actor == actor.actor)
    )

    return OrderedCollection(id="objects.json", items=items.all()).build()  # type: ignore


@router.get("/html/{actor_name}/o/{uuid}", response_class=HTMLResponse, tags=["html"])
@router.get(
    "/html/{actor_name}/{uuid}",
    response_class=HTMLResponse,
    deprecated=True,
    tags=["html"],
)
async def get_object_html(
    actor: PublishingActorForName,
    obj: PublishedObjectWithInteractionsForUUID,
    request: Request,
    should_serve: ShouldServe,
):
    """Returns the HTML representation for an object"""
    if actor.actor != obj.actor:
        raise HTTPException(404)

    if (
        ContentType.html not in should_serve
        and ContentType.activity_pub in should_serve
    ):
        return RedirectResponse(obj.data["id"])

    interactions = {
        str(interaction): [
            x.object_id for x in obj.interactions if x.interaction == interaction
        ]
        for interaction in [
            InteractionType.shares,
            InteractionType.likes,
            InteractionType.replies,
        ]
    }

    return templates.TemplateResponse(
        request,
        name="object.html.j2",
        context={
            "name": actor.name,
            "content": sanitize_html(obj.data["content"]),
            **interactions,
            "version": __version__,
        },
    )



from faststream import ExceptionMiddleware, Context

import logging
import json

from cattle_grid.account.account import account_for_actor
from cattle_grid.model.account import ErrorMessage
from cattle_grid.dependencies import AccountExchangePublisher, SqlSession
from cattle_grid.dependencies.processing import RoutingKey

logger = logging.getLogger(__name__)

exception_middleware = ExceptionMiddleware()


@exception_middleware.add_handler(Exception)
async def exception_handler(
    exception: Exception,
    routing_key: RoutingKey,
    session: SqlSession,
    publisher: AccountExchangePublisher,
    message=Context("message.body"),
):
    """When an exception occurs in processing, this handler will create
    an appropriate entry in `receive.NAME.error` in the account exchange"""
    try:
        data = json.loads(message)
        actor_id = data["actor"]
    except Exception:
        logger.exception(exception)
        logger.exception(message)
        return

    logger.info("Exception for %s", actor_id)
    logger.debug(exception, exc_info=True)
    account = await account_for_actor(session, actor_id)
    if not account:
        return
    name = account.name

    logger.info("Processing error occurred in exchange for account %s", name)

    await publisher(
        ErrorMessage(message=str(exception).split("\n"), routing_key=routing_key),
        routing_key=f"error.{name}",
    )



from dataclasses import dataclass, field
from typing import Any
from uuid import UUID

from bovine.activitystreams.utils import as_list
from uuid6 import uuid7


from .config import HtmlDisplayConfiguration
from .database import PublishingActor


@dataclass
class Publisher:
    """Class for manipulating objects being published"""

    actor: PublishingActor
    config: HtmlDisplayConfiguration
    obj: dict[str, Any]
    uuid: UUID = field(default_factory=uuid7)

    def __post_init__(self):
        if "id" in self.obj:
            obj_id = self.obj["id"]
            self.uuid = UUID(obj_id.split("/")[-1])
        else:
            self.obj = {
                **self.obj,
                "id": self.config.url_start(self.actor.actor) + str(self.uuid),
            }

    @property
    def object_for_store(self):
        return self.obj

    @property
    def object_for_remote(self):
        copy = {**self.obj}
        copy.update(self._collection_links())
        return self._add_url_to_obj(copy)

    def _collection_links(self):
        object_id = self.obj["id"]
        return {
            collection: f"{object_id}/{collection}"
            for collection in ["replies", "shares", "likes"]
        }

    def _add_url_to_obj(self, obj: dict):
        url_list = as_list(obj.get("url", []))

        url = (
            self.config.html_url_start(self.actor.actor)
            + self.actor.name
            + "/o/"
            + str(self.uuid)
        )

        obj["url"] = url_list + [self.create_html_link(url)]

        return obj

    def create_html_link(self, url: str | None = None):
        if url is None:
            url = self.actor.actor_id + "#html"
        return {"type": "Link", "mediaType": "text/html", "href": url}



"""
These are annotations useable when building a transformer in an extension,
see [here](../extensions/index.md#transformers).

Furthermore, you may use the annotations from [muck_out.cattle_grid][] starting
with Fetch or Transformed.
"""

from typing import Annotated
from fast_depends import Depends
from sqlalchemy import select
from cattle_grid.database.activity_pub_actor import Actor
from cattle_grid.dependencies import SqlSession


async def actor_for_transformer(session: SqlSession, actor_id: str) -> Actor:
    actor = await session.scalar(select(Actor).where(Actor.actor_id == actor_id))

    if actor is None:
        raise Exception("Actor not found")

    return actor


ActorTransforming = Annotated[Actor, Depends(actor_for_transformer)]
"""The actor in whose scope the transformation is running"""



from cattle_grid.database.activity_pub_actor import Blocking
from .testing import *  # noqa


def test_get_object_not_found(test_client):
    response = test_client.get(
        "/html_display/object/2fd16a00-309b-4f3a-9d91-aa9516e59c1f"
    )
    assert response.status_code == 404


def test_get_object(test_client, published_object):
    assert published_object

    object_id = published_object["id"]

    response = test_client.get(
        object_id,
        headers={
            "x-ap-location": object_id,
            "x-cattle-grid-requester": "http://remote.test/actor",
        },
    )

    assert response.status_code == 200
    assert response.headers["content-type"] == "application/activity+json"
    data = response.json()

    assert data["replies"] == object_id + "/replies"


def test_get_object_wrong_location(test_client, published_object):
    assert published_object

    object_id = published_object["id"]

    response = test_client.get(
        object_id,
        headers={
            "x-ap-location": "http://wrong.test",
            "x-cattle-grid-requester": "http://remote.test/actor",
        },
    )

    assert response.status_code == 404


def test_get_object_unauthorized(test_client, published_object):
    assert published_object

    object_id = published_object["id"]

    response = test_client.get(
        object_id,
        headers={
            "x-ap-location": object_id,
        },
    )

    assert response.status_code == 401


async def test_get_object_blocked(
    test_client, published_object, actor_for_test, sql_session
):
    remote_actor = "http://remote.test/actor"

    object_id = published_object["id"]
    sql_session.add(
        Blocking(
            actor=actor_for_test,
            blocking=remote_actor,
            active=True,
            request=actor_for_test.actor_id + "#block",
        )
    )
    await sql_session.commit()

    response = test_client.get(
        object_id,
        headers={
            "x-ap-location": object_id,
            "x-cattle-grid-requester": remote_actor,
        },
    )

    assert response.status_code == 401



from unittest.mock import AsyncMock
from fastapi import FastAPI
from fastapi.testclient import TestClient
import pytest

from .http_util import ContentType
from . import ShouldServe


@pytest.mark.parametrize(
    "headers,expected",
    [
        ({}, [ContentType.other]),
        ({"accept": "text/html"}, [ContentType.html]),
        ({"accept": "text/html; charsed=utf-8"}, [ContentType.html]),
        ({"accept": "application/activity+json"}, [ContentType.activity_pub]),
    ],
)
async def test_should_serve(headers, expected):
    mock = AsyncMock()
    app = FastAPI()

    @app.get("/")
    async def serve(should_serve: ShouldServe):
        await mock(should_serve)

    client = TestClient(app, headers=headers)

    client.get("/")
    mock.assert_awaited_once_with(expected)



"""Dependencies injected by fast_depends

cattle_grid uses dependencies to manage objects, one needs access to.
This works by declaring them using [fast_depends.Depends][] and then
injecting them using [fast_depends.inject][].

For example if you want to make a webrequest using the
[aiohttp.ClientSession][], you could use

```python
from cattle_grid.dependencies import ClientSession

async def web_request(session: ClientSession):
    response = await session.get("...")
```

This function can then be called via

```python
from fast_depends import inject

await inject(web_request)()
```

This package contains annotations that should be available in all code
using cattle_grid, i.e. extensions. The sub packages contain methods
for more specific use cases.
"""

import json
import aiohttp
import logging

from typing import Annotated, Callable
from dynaconf import Dynaconf
from fast_depends import Depends
from faststream import Context
from faststream.rabbit import RabbitExchange, RabbitBroker

from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession

from .internals import InternalExchange, CorrelationId
from .globals import get_engine, global_container


logger = logging.getLogger(__name__)


async def get_client_session():
    yield global_container.session


ClientSession = Annotated[aiohttp.ClientSession, Depends(get_client_session)]
"""The [aiohttp.ClientSession][] used by the application"""

ActivityExchange = Annotated[RabbitExchange, Depends(global_container.get_exchange)]
"""The activity exchange"""

AccountExchange = Annotated[
    RabbitExchange, Depends(global_container.get_account_exchange)
]
"""The account exchange"""

SqlAsyncEngine = Annotated[AsyncEngine, Depends(get_engine)]
"""Returns the SqlAlchemy AsyncEngine"""


SqlSessionMaker = Annotated[
    Callable[[], AsyncSession], Depends(global_container.get_session_maker)
]


async def with_sql_session(
    sql_session_maker=Depends(global_container.get_session_maker),
):
    async with sql_session_maker() as session:
        yield session


SqlSession = Annotated[AsyncSession, Depends(with_sql_session)]
"""SQL session that does not commit afterwards"""


async def with_session_commit(session: SqlSession):
    yield session
    await session.commit()


CommittingSession = Annotated[AsyncSession, Depends(with_session_commit)]
"""Session that commits the transaction"""

Config = Annotated[Dynaconf, Depends(global_container.get_config)]
"""Returns the configuration"""


def activity_exchange_publisher(
    correlation_id: CorrelationId,
    exchange: ActivityExchange,
    broker: RabbitBroker = Context(),
):
    async def publish(*args, **kwargs):
        kwargs_updated = {**kwargs}
        kwargs_updated.update(dict(exchange=exchange, correlation_id=correlation_id))

        return await broker.publish(*args, **kwargs_updated)

    return publish


def internal_exchange_publisher(
    correlation_id: CorrelationId,
    exchange: InternalExchange,
    broker: RabbitBroker = Context(),
):
    async def publish(*args, **kwargs):
        kwargs_updated = {**kwargs}
        kwargs_updated.update(dict(exchange=exchange, correlation_id=correlation_id))
        return await broker.publish(*args, **kwargs_updated)

    return publish


def internal_exchange_requester(
    correlation_id: CorrelationId,
    exchange: InternalExchange,
    broker: RabbitBroker = Context(),
):
    async def publish(*args, **kwargs):
        kwargs_updated = {**kwargs}
        kwargs_updated.update(dict(exchange=exchange, correlation_id=correlation_id))
        result = await broker.request(*args, **kwargs_updated)
        return json.loads(result.body)

    return publish


def account_exchange_publisher(
    correlation_id: CorrelationId,
    exchange: AccountExchange,
    broker: RabbitBroker = Context(),
):
    async def publish(*args, **kwargs):
        kwargs_updated = {**kwargs}
        kwargs_updated.update(dict(exchange=exchange, correlation_id=correlation_id))
        return await broker.publish(*args, **kwargs_updated)

    return publish


AccountExchangePublisher = Annotated[Callable, Depends(account_exchange_publisher)]
"""Publishes a message to the activity exchange"""

InternalExchangePublisher = Annotated[Callable, Depends(internal_exchange_publisher)]
"""Publishes a message to the internal exchange"""

InternalExchangeRequester = Annotated[Callable, Depends(internal_exchange_requester)]
"""Request a message to the internal exchange"""

ActivityExchangePublisher = Annotated[Callable, Depends(activity_exchange_publisher)]
"""Publishes a message to the activity exchange"""



from enum import StrEnum, auto
from pydantic import Field
from cattle_grid.model.common import WithActor


class NameActorMessage(WithActor):
    """Message for renaming an actor"""

    name: str = Field(description="Name for the actor", examples=["john"])


class ExportTokenResponse(WithActor):
    """Message containing the export information"""

    token: str = Field(description="One time token")
    export_url: str = Field(
        description="Url including the token the export is located at"
    )


class InteractionType(StrEnum):
    replies = auto()
    likes = auto()
    shares = auto()



from .types import Relationship, RelationshipStatus


def test_serialization():
    r = Relationship(status=RelationshipStatus.accepted)

    assert r.model_dump_json() == """{"status":"accepted","requests":[]}"""



from cattle_grid.extensions import Extension
from cattle_grid.extensions.util import lifespan_for_sql_alchemy_base_class

from .config import HtmlDisplayConfiguration
from .database import Base

extension = Extension(
    name="simple html display",
    module=__name__,
    lifespan=lifespan_for_sql_alchemy_base_class(Base),
    config_class=HtmlDisplayConfiguration,
)
extension.rewrite_group_name = "html_display"
extension.rewrite_rules = {"publish_object": "html_display_publish_object"}



from typing import Annotated, Awaitable, Callable

from fast_depends import Depends
from faststream import Context
from faststream.rabbit import RabbitExchange

from cattle_grid.config.rewrite import RewriteConfiguration
from cattle_grid.model.extension import MethodInformationModel
from cattle_grid.model.lookup import LookupMethod
from .globals import (
    get_transformer,
    get_lookup,
    global_container,
    get_method_information,
)


Transformer = Annotated[Callable[..., Awaitable[dict]], Depends(get_transformer)]
"""The transformer loaded from extensions"""

LookupAnnotation = Annotated[LookupMethod, Depends(get_lookup)]
"""The lookup method loaded from extensions"""


InternalExchange = Annotated[
    RabbitExchange, Depends(global_container.get_internal_exchange)
]
"""The interal activity exchange"""

CorrelationId = Annotated[str, Context("message.correlation_id")]
"""The correlation id of the message"""

MethodInformation = Annotated[
    list[MethodInformationModel], Depends(get_method_information)
]
"""Returns the information about the methods that are a part of the exchange"""


RewriteRules = Annotated[
    RewriteConfiguration, Depends(global_container.get_rewrite_rules)
]
"""Rewturns the rewrite configuration"""



"""

<div class="grid cards" markdown>

- [:material-api:{ .lg .middle } __OpenAPI__](../assets/redoc.html?url=openapi_simple_html_display.json)
- [:material-file-code:{ .lg .middle } __Source__](https://codeberg.org/bovine/cattle_grid/src/branch/main/cattle_grid/extensions/examples/html_display)

</div>


The goal of this extension is to illustrate the mechanism how cattle_grid
can be used to create a webpage displaying the public posts of its users.
The focus is on providing a working solution to key problems than on
providing the optimal solution.

An example of a feed is available at [@release](https://dev.bovine.social/@release).
The current scope is:

- Display actor and posts
- Correct redirecting behavior for easy use
- Display replies (todo)
- Export
- Import from compatible sources (todo)

This is not an independent solution. It relies on both the [context](./context.md) and
[simple_storage](./simple_storage.md) extension of cattle_grid.
"""

import json
import logging
from uuid import uuid4

from bovine.activitystreams.utils import is_public
from sqlalchemy import select

from muck_out.cattle_grid import ParsedActivity, ParsedEmbeddedObject
from uuid6 import uuid7

from cattle_grid.account.account import account_for_actor
from cattle_grid.dependencies import (
    AccountExchangePublisher,
    ActivityExchangePublisher,
    CommittingSession,
    SqlSession,
)
from cattle_grid.dependencies.processing import FactoriesForActor

from cattle_grid.extensions.examples.html_display.storage import object_for_object_id
from cattle_grid.manage.actor import ActorManager
from cattle_grid.model import ActivityMessage
from cattle_grid.model.exchange import UpdateActorMessage
from cattle_grid.model.exchange_update_actor import UpdateActionType, UpdateUrlAction
from cattle_grid.activity_pub.activity import actor_deletes_themselves

from .dependencies import PublishingActor
from .publisher import Publisher

from .database import (
    ExportPermission,
    PublishedObject,
    PublishedObjectInteraction,
    PublishingActor as DBPublishingActor,
)
from .router import router
from .types import ExportTokenResponse, InteractionType, NameActorMessage
from .extension_declaration import extension

logger = logging.getLogger(__name__)


extension.include_router(router)


@extension.subscribe("html_display_publish_object")
async def html_publish_object(
    message: ActivityMessage,
    session: CommittingSession,
    actor: PublishingActor,
    config: extension.Config,  # type:ignore
    factories: FactoriesForActor,
    activity_publisher: ActivityExchangePublisher,
):
    """Publishes an object"""
    obj = message.data

    if not is_public(obj):
        await activity_publisher(
            ActivityMessage(actor=message.actor, data=obj),
            routing_key="publish_object",
        )
        return

    if obj.get("id"):
        raise ValueError("Object ID must not be set")

    if obj.get("attributedTo") != message.actor:
        raise ValueError("Actor must match object attributedTo")

    publisher = Publisher(actor, config, obj)
    session.add(
        PublishedObject(
            id=publisher.uuid, data=publisher.object_for_store, actor=actor.actor
        )
    )

    activity = factories[0].create(publisher.object_for_remote).build()

    await activity_publisher(
        ActivityMessage(actor=message.actor, data=activity),
        routing_key="publish_activity",
    )


@extension.subscribe("html_display_name")
async def name_actor(
    message: NameActorMessage,
    actor: PublishingActor,
    session: CommittingSession,
    activity_publisher: ActivityExchangePublisher,
    config: extension.Config,  # type:ignore
):
    """Sets the display name of the actor"""
    if message.actor != actor.actor:
        raise Exception("Actor mismatch")

    actor.name = message.name

    await activity_publisher(
        UpdateActorMessage(
            actor=actor.actor,
            actions=[
                UpdateUrlAction(
                    action=UpdateActionType.add_url,
                    url=config.html_url(actor.actor, actor.name),
                    media_type="text/html",
                )
            ],
        ),
        routing_key="update_actor",
    )

    if config.automatically_add_users_to_group:
        manager = ActorManager(actor_id=actor.actor, session=session)
        await manager.add_to_group("html_display")


@extension.subscribe("html_display_export", replies=True)
async def export_create_token(
    actor: PublishingActor,
    session: CommittingSession,
    publisher: AccountExchangePublisher,
    config: extension.Config,  # type:ignore
):
    """Provides a token and download link for the export of stored objects

    The response is of type [cattle_grid.extensions.examples.html_display.types.ExportTokenResponse][] and
    send to `receive.NAME.response.trigger`."""
    token_uuid = uuid4()
    session.add(ExportPermission(publishing_actor=actor, one_time_token=token_uuid))

    account = await account_for_actor(session, actor.actor)

    if not account:
        raise Exception("Could not find account")

    await publisher(
        ExportTokenResponse(
            actor=actor.actor,
            token=str(token_uuid),
            export_url=config.html_url(actor.actor, actor.name)
            + f"/export?token={str(token_uuid)}",
        ),
        routing_key=f"receive.{account.name}.response.trigger",
    )


@extension.subscribe("outgoing.Delete")
async def outgoing_delete(
    message: ActivityMessage,
    session: SqlSession,
):
    """Deletes the publishing actor, if they delete themself"""
    activity = message.data.get("raw")
    if not isinstance(activity, dict):
        return
    if not actor_deletes_themselves(activity):
        return

    actor = await session.scalar(
        select(DBPublishingActor).where(DBPublishingActor.actor == message.actor)
    )
    if not actor:
        return

    logger.info("Deleting publishing actor with name %s", actor.name)

    await session.delete(actor)
    await session.flush()


@extension.subscribe("incoming.Like")
async def html_display_incoming_like(
    activity: ParsedActivity,
    message: ActivityMessage,
    session: CommittingSession,
):
    """Handles likes"""
    if activity is None:
        logger.info("Unparsed like: " + json.dumps(message.data, indent=2))
        return
    object_id = activity.object

    if not isinstance(object_id, str):
        return

    published_object = await object_for_object_id(session, object_id)

    session.add(
        PublishedObjectInteraction(
            id=uuid7(),
            published_object=published_object,
            object_id=activity.id,
            interaction=InteractionType.likes,
        )
    )


@extension.subscribe("incoming.Announce")
async def html_display_incoming_announce(
    activity: ParsedActivity,
    message: ActivityMessage,
    session: CommittingSession,
):
    """Handles announces"""
    if activity is None:
        logger.info("Unparsed announce: " + json.dumps(message.data, indent=2))
        return
    object_id = activity.object

    if not isinstance(object_id, str):
        return

    published_object = await object_for_object_id(session, object_id)

    session.add(
        PublishedObjectInteraction(
            id=uuid7(),
            published_object=published_object,
            object_id=activity.id,
            interaction=InteractionType.shares,
        )
    )


@extension.subscribe("incoming.Create")
async def html_display_incoming_create(
    embedded_object: ParsedEmbeddedObject,
    session: CommittingSession,
):
    """Handles replies"""
    if not embedded_object:
        return

    object_id = embedded_object.in_reply_to

    if not isinstance(object_id, str):
        return

    published_object = await object_for_object_id(session, object_id)

    session.add(
        PublishedObjectInteraction(
            id=uuid7(),
            published_object=published_object,
            object_id=embedded_object.id,
            interaction=InteractionType.replies,
        )
    )



import pytest
from cattle_grid.extensions.load import build_transformer

from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.activity_pub_actor import Follower

from muck_out.extension import extension as muck_out
from . import extension


@pytest.fixture
def transformer():
    return build_transformer([muck_out, extension])


@pytest.fixture
def remote_actor_profile():
    return {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            "http://joinmastodon.org/ns",
            "https://w3id.org/security/v1",
        ],
        "discoverable": False,
        "featured": "http://gotosocial/users/cookie/collections/featured",
        "followers": "http://gotosocial/users/cookie/followers",
        "following": "http://gotosocial/users/cookie/following",
        "id": "http://gotosocial/users/cookie",
        "inbox": "http://gotosocial/users/cookie/inbox",
        "manuallyApprovesFollowers": True,
        "name": "cookie",
        "outbox": "http://gotosocial/users/cookie/outbox",
        "preferredUsername": "cookie",
        "publicKey": {
            "id": "http://gotosocial/users/cookie/main-key",
            "owner": "http://gotosocial/users/cookie",
            "publicKeyPem": "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0SwHxmzQTZrdQkiD1a68\nMg4gDbkqAzocxngdjXOG4Y92q2MZB/ByRaPS7ap4e5EKDa3TaVIz0JeoZGytlmMu\ntvBRKfLyutGZLz6hkrlkSGzQT9XGYrmfkrXqCjSq4kML0ncMBpaf+VQPjpiqcBTp\n0N3NZCNp3tfarGeL1+CNJpz6qZI0UsSp3pStrE1+DIQbIay4tlgHdjvX0nBPalen\nVWXxdSgywF1A4mEmdwkSVJ7Jl7pQEGT4q5ojFuAdTcTFMGskUmez8trVTqHNAs9M\n3TmZbtWgNKG/AuYrR1CPkEoup+aCPpjvMR4EWtnqxbTFfmN1DUGepPjVPHzb+UBQ\nRQIDAQAB\n-----END PUBLIC KEY-----\n",
        },
        "tag": [],
        "type": "Person",
        "url": "http://gotosocial/@cookie",
    }


async def test_transformer_with_actor_id(transformer, actor_for_test):
    result = await transformer({"raw": {}}, actor_id=actor_for_test.actor_id)

    assert "relationship" in result
    assert result["relationship"] == {}


async def test_transformer_with_follower(
    transformer, remote_actor_profile, actor_for_test, sql_session
):
    sql_session.add(
        Follower(
            actor=actor_for_test,
            follower=remote_actor_profile["id"],
            accepted=True,
            request=remote_actor_profile["id"] + "#follow_request",
        )
    )

    await sql_session.commit()

    result = await transformer(
        {"raw": remote_actor_profile}, actor_id=actor_for_test.actor_id
    )

    relationship = result["relationship"]

    assert "follower" in relationship



from .types import Relationship, RelationshipStatus


def determine_status(info: list[tuple[bool, str]]) -> Relationship:
    if len(info) == 0:
        return Relationship(status=RelationshipStatus.none)

    status = (
        RelationshipStatus.accepted
        if any(x[0] for x in info)
        else RelationshipStatus.waiting
    )
    requests = [x[1] for x in info]

    return Relationship(status=status, requests=requests)



from bovine.activitystreams import factories_for_actor_object
from muck_out import normalize_data
import pytest

from typing import Any

from cattle_grid.model import ActivityMessage
from cattle_grid.dependencies.globals import global_container


from .testing import *  # noqa


@pytest.mark.parametrize("interaction", ["likes", "shares", "replies"])
def test_interaction_collections(test_client, published_object, interaction):
    object_id = published_object["id"]
    response = test_client.get(
        object_id,
        headers={
            "x-ap-location": object_id,
            "x-cattle-grid-requester": "http://remote.test/actor",
        },
    )

    assert response.status_code == 200
    data = response.json()
    interaction_collection = data[interaction]

    response = test_client.get(
        interaction_collection,
        headers={
            "x-ap-location": interaction_collection,
            "x-cattle-grid-requester": "http://remote.test/actor",
        },
    )

    assert response.status_code == 200
    data = response.json()

    assert data["id"] == interaction_collection
    assert data["type"] == "OrderedCollection"


async def interact(remote_actor, test_broker, interaction: str, obj: dict[str, Any]):
    activity_factory, object_factory = factories_for_actor_object(
        {"id": remote_actor}, id_generator=lambda: f"{remote_actor}/id"
    )
    actor = obj.get("attributedTo", "broken")
    activity = {}
    routing_key = None

    match interaction:
        case "likes":
            activity = activity_factory.like(obj.get("id"), to={actor}).build()
            routing_key = "incoming.Like"
        case "shares":
            activity = activity_factory.announce(obj.get("id"), to={actor}).build()
            routing_key = "incoming.Announce"
        case "replies":
            reply = object_factory.reply(obj, content="oh a reply").build()
            activity = activity_factory.create(reply).build()
            routing_key = "incoming.Create"

    normalized = normalize_data(activity).model_dump()
    msg = ActivityMessage(actor=actor, data={"raw": activity, "parsed": normalized})

    await test_broker.publish(
        msg,
        routing_key=routing_key,
        exchange=global_container.exchange,
    )


@pytest.mark.parametrize("interaction", ["likes", "shares", "replies"])
async def test_interaction_collection_with_interaction(
    test_client,
    published_object,
    interaction,
    test_broker,
):
    remote_actor = "http://remote.test/actor"
    object_id = published_object["id"]
    response = test_client.get(
        object_id,
        headers={
            "x-ap-location": object_id,
            "x-cattle-grid-requester": remote_actor,
        },
    )

    assert response.status_code == 200
    data = response.json()
    interaction_collection = data[interaction]

    await interact(remote_actor, test_broker, interaction, data)

    response = test_client.get(
        interaction_collection,
        headers={
            "x-ap-location": interaction_collection,
            "x-cattle-grid-requester": "http://remote.test/actor",
        },
    )

    assert response.status_code == 200
    data = response.json()

    assert data["id"] == interaction_collection
    assert data["type"] == "OrderedCollection"
    assert data.get("orderedItems")



from sqlalchemy import select
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.account import ActorForAccount
from cattle_grid.model.exchange_update_actor import RenameActorAction

from .actor_update import handle_actor_action


async def test_handle_actor_action_rename(actor_with_account, sql_session):
    action = RenameActorAction(name="new name")

    await handle_actor_action(actor_with_account, sql_session, action)

    actor_for_account = await sql_session.scalar(
        select(ActorForAccount).where(
            ActorForAccount.actor == actor_with_account.actor_id
        )
    )
    assert actor_for_account.name == "new name"



from typing import Annotated
from unittest.mock import AsyncMock
from fast_depends import Depends
from faststream.rabbit import RabbitBroker
from faststream.rabbit.testing import TestRabbitBroker

from .internals import CorrelationId


async def test_correlation_id():
    broker = RabbitBroker()

    mock = AsyncMock()

    async def depends_on_correlation_id(correlation_id: CorrelationId):
        await mock(correlation_id)
        return "test"

    @broker.subscriber("test")
    async def my_func(var: Annotated[str, Depends(depends_on_correlation_id)]): ...

    async with TestRabbitBroker(broker) as tbr:
        await tbr.publish({}, "test", correlation_id="corr")

    mock.assert_awaited_once_with("corr")



"""
FastAPI uses [fastapi.Depends][] instead of [fast_depends.Depends][],
so when building a [fastapi.APIRouter][], one needs to use dependencies
using the former. These are provided in this package.
"""

import json
import aiohttp

from typing import Annotated, Callable, Awaitable, Dict, List
from dynaconf import Dynaconf
from faststream.rabbit import RabbitBroker, RabbitExchange

from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, AsyncSession

from fastapi import Depends

from cattle_grid.model.extension import MethodInformationModel
from .globals import (
    get_engine,
    get_transformer,
    get_method_information,
    global_container,
)

SqlAsyncEngine = Annotated[AsyncEngine, Depends(get_engine)]
"""Returns the SqlAlchemy AsyncEngine"""

Transformer = Annotated[Callable[[Dict], Awaitable[Dict]], Depends(get_transformer)]
"""The transformer loaded from extensions"""

Broker = Annotated[RabbitBroker, Depends(global_container.get_broker)]
"""The RabbitMQ broker"""
InternalExchange = Annotated[
    RabbitExchange, Depends(global_container.get_internal_exchange)
]

ActivityExchange = Annotated[RabbitExchange, Depends(global_container.get_exchange)]
"""The Activity Exchange"""


MethodInformation = Annotated[
    List[MethodInformationModel], Depends(get_method_information)
]
"""Returns the information about the methods that are a part of the exchange"""


async def with_fast_api_session(sql_engine: SqlAsyncEngine):
    async with async_sessionmaker(sql_engine)() as session:
        yield session


SqlSession = Annotated[AsyncSession, Depends(with_fast_api_session)]
"""Session annotation to be used with FastAPI"""


async def with_committing_sql_session(session: SqlSession):
    yield session
    await session.commit()


CommittingSqlSession = Annotated[AsyncSession, Depends(with_committing_sql_session)]
"""Session annotation to be used with FastAPI. A commit is performed, after processing the request"""


Config = Annotated[Dynaconf, Depends(global_container.get_config)]
"""Returns the configuration"""


async def get_client_session():
    yield global_container.session


ClientSession = Annotated[aiohttp.ClientSession, Depends(get_client_session)]
"""The [aiohttp.ClientSession][] used by the application"""


def activity_exchange_publisher(
    exchange: ActivityExchange,
    broker: Broker,
):
    async def publish(*args, **kwargs):
        kwargs_updated = {**kwargs}
        kwargs_updated.update(dict(exchange=exchange))
        return await broker.publish(*args, **kwargs_updated)

    return publish


ActivityExchangePublisher = Annotated[Callable, Depends(activity_exchange_publisher)]
"""Publisher to the activity exchange"""


def activity_exchange_requester(
    exchange: ActivityExchange,
    broker: Broker,
):
    async def request(*args, **kwargs):
        kwargs_updated = {**kwargs}
        kwargs_updated.update(dict(exchange=exchange))
        result = await broker.request(*args, **kwargs_updated)
        return json.loads(result.body)

    return request


ActivityExchangeRequester = Annotated[Callable, Depends(activity_exchange_requester)]
"""Requester to the activity exchange"""



import json

from cattle_grid.model.exchange import UpdateActorMessage


model_name_to_model = {"UpdateActorMessage": UpdateActorMessage}


def json_schema_for_model(model_name: str):
    model = model_name_to_model[model_name]

    schema = model.model_json_schema()
    print(json.dumps(schema, indent=2))



import datetime

from sqlalchemy import Enum, ForeignKey, String, func, JSON
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from sqlalchemy_utils.types import UUIDType

from cattle_grid.extensions.examples.html_display.types import InteractionType


class Base(AsyncAttrs, DeclarativeBase):
    """Base model"""

    pass


class PublishingActor(Base):
    __tablename__ = "html_display_publishing_actor"

    id: Mapped[int] = mapped_column(primary_key=True)

    actor: Mapped[str] = mapped_column(String(256))
    name: Mapped[str] = mapped_column(String(256), unique=True)


class PublishedObject(Base):
    """HTML display object in the database"""

    __tablename__ = "html_display_stored_object"

    id: Mapped[bytes] = mapped_column(UUIDType(binary=True), primary_key=True)
    data: Mapped[dict] = mapped_column(JSON)
    actor: Mapped[str] = mapped_column()
    create_date: Mapped[datetime.datetime] = mapped_column(server_default=func.now())

    interactions: Mapped[list["PublishedObjectInteraction"]] = relationship(
        viewonly=True
    )


class PublishedObjectInteraction(Base):
    """HTML display object in the database"""

    __tablename__ = "html_display_stored_object_interaction"

    id: Mapped[bytes] = mapped_column(UUIDType(binary=True), primary_key=True)

    published_object_id: Mapped[str] = mapped_column(
        ForeignKey("html_display_stored_object.id")
    )
    published_object: Mapped[PublishedObject] = relationship()

    object_id: Mapped[str] = mapped_column()

    interaction: Mapped[InteractionType] = mapped_column(Enum(InteractionType))


class ExportPermission(Base):
    """Records a one time token required to download the export"""

    __tablename__ = "html_display_export_permission"

    id: Mapped[int] = mapped_column(primary_key=True)
    publishing_actor_id: Mapped[str] = mapped_column(
        ForeignKey("html_display_publishing_actor.id")
    )
    publishing_actor: Mapped[PublishingActor] = relationship()
    one_time_token: Mapped[bytes] = mapped_column(UUIDType(binary=True))



from .database import PublishingActor
from .storage import object_for_object_id, publishing_actor_for_actor_id

from .testing import *  # noqa


async def test_publishing_actor_for_actor_id(sql_session):
    actor_id = "http://actor.test/some/id"

    result = await publishing_actor_for_actor_id(sql_session, actor_id)

    assert isinstance(result, PublishingActor)
    assert result.actor == actor_id


async def test_publishing_actor_for_actor_id_returns_stored_actor(sql_session):
    actor_id = "http://actor.test/some/id"

    one = await publishing_actor_for_actor_id(sql_session, actor_id)

    await sql_session.commit()

    two = await publishing_actor_for_actor_id(sql_session, actor_id)

    assert one.name == two.name


async def test_object_for_object_id_unknown(sql_session):
    object_id = "http://object.test/some/id"

    result = await object_for_object_id(sql_session, object_id)

    assert result is None


async def test_object_for_object_id(sql_session, published_object):
    object_id = published_object.get("id")

    result = await object_for_object_id(sql_session, object_id)

    assert result



import logging
from typing import Any

from fast_depends import inject
from sqlalchemy import select, func
from sqlalchemy.ext.asyncio import AsyncSession

from bovine.activitystreams.utils import id_for_object

from cattle_grid.dependencies.internals import Transformer
from cattle_grid.model import ActivityMessage
from cattle_grid.database.activity_pub_actor import Actor, Blocking

from cattle_grid.dependencies import (
    AccountExchangePublisher,
    ActivityExchangePublisher,
    SqlSession,
)


from cattle_grid.database.account import ActorForAccount
from cattle_grid.model.account import EventInformation, EventType

logger = logging.getLogger(__name__)


async def should_shovel_activity(session: AsyncSession, activity: dict) -> bool:
    """Some activities like Block or Undo Block should not be visible to the user. This method
    returns False if this is the case."""

    activity_type = activity.get("type")

    if activity_type == "Block":
        return False

    if activity_type == "Undo":
        object_id = id_for_object(activity.get("object"))
        blocking = await session.scalar(
            func.count(
                select(Blocking.id)
                .where(Blocking.request == object_id)
                .scalar_subquery()
            )
        )

        if blocking:
            return False

    return True


@inject
async def shovel(
    actor: str,
    activity: dict[str, Any],
    session: SqlSession,
    transformer: Transformer,
    publisher: ActivityExchangePublisher,
    account_publisher: AccountExchangePublisher,
    direction: EventType,
):
    actor_for_account = await session.scalar(
        select(ActorForAccount).where(ActorForAccount.actor == actor)
    )
    if actor_for_account is None:
        logger.warning("Got actor without account %s", actor)
        return

    account_name = actor_for_account.account.name
    to_shovel = await transformer({"raw": activity}, actor_id=actor)
    activity_type = activity.get("type")

    event_info = EventInformation(
        actor=actor,
        event_type=direction,
        data=to_shovel,
    )

    await account_publisher(
        event_info,
        routing_key=f"receive.{account_name}.{str(direction)}",
    )
    await account_publisher(
        event_info,
        routing_key=f"receive.{account_name}.{str(direction)}.{activity_type}",
    )
    await publisher(
        ActivityMessage(actor=actor, data=to_shovel),
        routing_key=f"{str(direction)}.{activity_type}",
    )


async def incoming_shovel(
    msg: ActivityMessage,
    session: SqlSession,
) -> None:
    """Transfers the message from the RawExchange to the
    Activity- and Account one.

    The message is passed through the transformer.
    """
    logger.info("incoming shovel")

    if not await should_shovel_activity(session, msg.data):
        return

    # FIXME: Use join to combine the next two queries ...

    db_actor = await session.scalar(select(Actor).where(Actor.actor_id == msg.actor))
    if not db_actor:
        raise ValueError("Actor not found in database")

    blocking = await session.scalar(
        select(Blocking)
        .where(Blocking.actor == db_actor)
        .where(Blocking.blocking == msg.data.get("actor"))
        .where(Blocking.active)
    )
    if blocking:
        return

    return await shovel(
        msg.actor,
        msg.data,
        direction=EventType.incoming,
    )  # type: ignore


async def outgoing_shovel(msg: ActivityMessage) -> None:
    return await shovel(
        msg.actor,
        msg.data,
        direction=EventType.outgoing,
    )  # type: ignore



from cattle_grid.account.account import group_names_for_actor
from cattle_grid.model.account import TriggerMessage
from cattle_grid.dependencies import ActivityExchangePublisher, SqlSession
from cattle_grid.dependencies.internals import RewriteRules

from .annotations import ActorForAccountFromMessage, MethodFromRoutingKey


async def handle_trigger(
    msg: TriggerMessage,
    actor: ActorForAccountFromMessage,
    method: MethodFromRoutingKey,
    session: SqlSession,
    rewrite_rules: RewriteRules,
    publisher: ActivityExchangePublisher,
):
    """Used to trigger a method performed by the actor.

    The main thing an actor can do is send activities to
    the Fediverse. This can be done with `send_message`.
    This can be extended in cattle_grid through extensions.

    However the methods to update the actor profile and delete
    the actor are also called via a trigger.
    """

    if rewrite_rules:
        group_names = await group_names_for_actor(session, actor)
        method = rewrite_rules.rewrite(method, group_names)

    await publisher(msg, routing_key=method)



import logging
from typing import Any

from sqlalchemy import select
from cattle_grid.database.activity_pub_actor import Blocking, Follower, Following
from cattle_grid.dependencies import SqlSession
from cattle_grid.dependencies.transformer import ActorTransforming
from cattle_grid.extensions import Extension

from muck_out.cattle_grid import TransformedActor

from cattle_grid.extensions.examples.relationship.status import determine_status

extension = Extension("relationship", __name__)

logger = logging.getLogger(__name__)


@extension.transform(inputs=["parsed"], outputs=["releationship"])
async def transform(
    transformed_actor: TransformedActor, actor: ActorTransforming, session: SqlSession
) -> dict[str, Any]:
    if transformed_actor is None:
        return {"relationship": {}}

    logger.info("Processing relationships")

    result = {}

    follower = await session.execute(
        select(Follower.accepted, Follower.request)
        .where(Follower.actor == actor)
        .where(Follower.follower == transformed_actor.id)
    )
    following = await session.execute(
        select(Following.accepted, Following.request)
        .where(Following.actor == actor)
        .where(Following.following == transformed_actor.id)
    )
    blocking = await session.execute(
        select(Blocking.active, Blocking.request)
        .where(Blocking.actor == actor)
        .where(Blocking.blocking == transformed_actor.id)
    )

    result["follower"] = determine_status(follower.all()).model_dump()  # type: ignore
    result["following"] = determine_status(following.all()).model_dump()  # type: ignore
    result["blocking"] = determine_status(blocking.all()).model_dump()  # type: ignore

    return {"relationship": result}



import logging

from cattle_grid.activity_pub.enqueuer import determine_activity_type
from cattle_grid.activity_pub.processing.common import MessageBovineActor
from cattle_grid.dependencies import (
    InternalExchangePublisher,
    InternalExchangeRequester,
)
from cattle_grid.dependencies.internals import LookupAnnotation, Transformer
from cattle_grid.model import ActivityMessage, FetchMessage


logger = logging.getLogger(__name__)


async def send_message(
    msg: ActivityMessage,
    publisher: InternalExchangePublisher,
) -> None:
    """Takes a message and ensure it is distributed appropriately"""

    content = msg.data
    activity_type = determine_activity_type(content)

    if not activity_type:
        return

    to_send = ActivityMessage(actor=msg.actor, data=content)

    await publisher(
        to_send,
        routing_key=f"outgoing.{activity_type}",
    )


async def fetch_object(msg: FetchMessage, requester: InternalExchangeRequester) -> dict:
    """Used to fetch an object as an RPC method"""
    result = await requester(
        msg,
        routing_key="fetch_object",
    )
    return result


async def fetch(
    message: FetchMessage,
    transformer: Transformer,
    lookup: LookupAnnotation,
    actor: MessageBovineActor,
) -> dict:
    """Used to fetch an object as an RPC method. In difference to `fetch_object`,
    this method applies the transformer."""

    from cattle_grid.activity_pub.processing.remote import fetch_object

    logger.info("fetch started")
    result = await fetch_object(
        message,
        actor,
        lookup,
    )
    logger.info("fetch done", result)

    transformed = await transformer({"raw": result}, actor_id=message.actor)
    return transformed



from enum import StrEnum, auto
from pydantic import BaseModel, Field


class RelationshipStatus(StrEnum):
    """The (technical) states of a relationship (follower, following, blocking)"""

    none = auto()
    accepted = auto()
    waiting = auto()


class Relationship(BaseModel):
    """Description of the relationship of an actor to another one"""

    status: RelationshipStatus = Field(
        description="The current status of the relationship"
    )
    requests: list[str] = Field(
        default=[], description="List of URIs of the involved requests"
    )



import asyncio
from behave import given, then, when

from cattle_grid.extensions.examples.html_display.types import NameActorMessage


@given('"{alice}" sets her display name to "{alex}"')  # type: ignore
async def html_set_display_name(context, alice, alex):
    connection = context.connections[alice]
    alice_id = context.actors[alice].get("id")

    await connection.trigger(
        "html_display_name", NameActorMessage(actor=alice_id, name=alex).model_dump()
    )

    await asyncio.sleep(0.2)


@then("The profile contains an url")  # type: ignore
def profile_contains_url(context):
    urls = context.profile.get("url")

    assert isinstance(urls, list)
    assert len(urls) >= 1


@when('"{alice}" requests an export token')  # type: ignore
async def request_token(context, alice):
    connection = context.connections[alice]
    alice_id = context.actors[alice].get("id")

    context.export_response = await connection.trigger(
        "html_display_export", {"actor": alice_id}
    )

    assert "token" in context.export_response


@when("The export url is retrieved")  # type: ignore
async def step_impl(context):
    async with context.session.get(
        context.export_response["export_url"],
    ) as response:
        context.response = response
        context.response_content = await response.text()


@then('The response contains "{text}"')  # type: ignore
async def response_contains(context, text):
    content = context.response_content
    assert text in content, f"Got content: {content}"


@then('A request to the object using accept "{media_type}" gets redirected')  # type: ignore
async def is_redirected(context, media_type):
    async with context.session.get(
        context.media_type_link,
        headers={"accept": media_type},
        allow_redirects=False,
    ) as response:
        assert response.status >= 300 and response.status < 400, (
            f"{response.status} is not a redirect"
        )



from typing import Annotated

from fast_depends import Depends
from cattle_grid.dependencies.processing import RoutingKey


def name_from_routing_key(
    routing_key: RoutingKey,
) -> str:
    """
    ```pycon
    >>> name_from_routing_key("receiving.alice")
    'alice'

    >>> name_from_routing_key("receiving.alice.action.fetch")
    'alice'

    ```
    """
    return routing_key.split(".")[1]


AccountName = Annotated[str, Depends(name_from_routing_key)]
"""Assigns the account name extracted from the routing key"""



from uuid import uuid4
import pytest

from cattle_grid.extensions.examples.html_display.database import ExportPermission

from .storage import publishing_actor_for_actor_id

from .testing import *  # noqa


def test_get_object_html(test_client, published_object):
    assert published_object
    url_in_obj = published_object.get("url", [])
    url = url_in_obj[0].get("href").replace("@", "html_display/html/")

    response = test_client.get(url)

    assert response.status_code == 200
    assert response.headers["content-type"].startswith("text/html")


async def test_get_object_html_not_found(test_client, publishing_actor_for_test):
    response = test_client.get(
        f"/html_display/html/{publishing_actor_for_test.name}/2fd16a00-309b-4f3a-9d91-aa9516e59c1f"
    )

    assert response.status_code == 404


async def test_get_object_html_actor_mismatch(
    sql_session, test_client, published_object
):
    actor = await publishing_actor_for_actor_id(sql_session, "some_id")
    await sql_session.commit()

    response = test_client.get(
        f"/html_display/html/{actor.name}/" + published_object.get("id").split("/")[-1]
    )

    assert response.status_code == 404


@pytest.mark.parametrize(
    "query,status",
    [
        ("", 422),
        ("?token=test", 422),
        ("?token=f47ac10b-58cc-4372-a567-0e02b2c3d479", 401),
    ],
)
async def test_export_objects_unauthorized(
    publishing_actor_for_test, test_client, query, status
):
    response = test_client.get(
        f"/html_display/html/{publishing_actor_for_test.name}/export{query}"
    )

    assert response.status_code == status


async def test_export_objects_authorized(
    sql_session, test_client, publishing_actor_for_test
):
    uuid = uuid4()
    sql_session.add(
        ExportPermission(
            publishing_actor=publishing_actor_for_test, one_time_token=uuid
        )
    )

    await sql_session.commit()

    response = test_client.get(
        f"/html_display/html/{publishing_actor_for_test.name}/export?token={uuid}"
    )

    assert response.status_code == 200
    data = response.json()

    assert data["type"] == "OrderedCollection"


async def test_export_objects_with_object(
    sql_session, test_client, publishing_actor_for_test, published_object
):
    uuid = uuid4()
    sql_session.add(
        ExportPermission(
            publishing_actor=publishing_actor_for_test, one_time_token=uuid
        )
    )

    await sql_session.commit()

    response = test_client.get(
        f"/html_display/html/{publishing_actor_for_test.name}/export?token={uuid}"
    )

    assert response.status_code == 200
    data = response.json()

    assert data["type"] == "OrderedCollection"

    assert data["totalItems"] == 1
    assert len(data["orderedItems"]) == 1



import pytest

from unittest.mock import AsyncMock
from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from cattle_grid.model import ActivityMessage
from cattle_grid.activity_pub.actor import create_actor
from cattle_grid.database.activity_pub_actor import Blocking
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.dependencies.globals import global_container

from .shovel import should_shovel_activity
from . import create_router


@pytest.fixture
def mock_incoming():
    return AsyncMock()


@pytest.fixture
def mock_outgoing():
    return AsyncMock()


async def fake_transformer(a, **kwargs):
    return a


@pytest.fixture
async def test_broker(mock_incoming, mock_outgoing):
    broker = RabbitBroker()
    broker.include_router(create_router())

    broker.subscriber(
        RabbitQueue("incoming", routing_key="incoming.*"),
        global_container.exchange,
    )(mock_incoming)
    broker.subscriber(
        RabbitQueue("outgoing", routing_key="outgoing.*"),
        global_container.exchange,
    )(mock_outgoing)

    global_container.transformer = fake_transformer

    async with TestRabbitBroker(broker) as tbr:
        yield tbr


async def test_incoming_message(test_broker, mock_incoming, actor_with_account):
    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": "http://remote.test/actor",
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor_with_account.actor_id,
        data=activity_pub,
    )

    await test_broker.publish(
        activity,
        exchange=global_container.internal_exchange,
        routing_key="incoming.Activity",
    )

    mock_incoming.assert_awaited_once()

    args = mock_incoming.call_args[1]
    result = args["data"]

    assert result["raw"]["@context"] == activity_pub["@context"]


async def test_incoming_block(test_broker, mock_incoming, actor_with_account):
    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Block",
        "actor": "http://remote.test/actor",
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(actor=actor_with_account.actor_id, data=activity_pub)

    await test_broker.publish(
        activity,
        exchange=global_container.internal_exchange,
        routing_key="incoming.Activity",
    )

    mock_incoming.assert_not_awaited()


async def test_incoming_from_blocked_user(
    test_broker, mock_incoming, sql_session, actor_with_account
):
    remote_actor = "http://remote.test/actor"

    sql_session.add(
        Blocking(
            actor=actor_with_account,
            blocking=remote_actor,
            request="http://blocked.test/id",
            active=True,
        )
    )
    await sql_session.commit()

    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": remote_actor,
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor_with_account.actor_id,
        data=activity_pub,
    )

    await test_broker.publish(
        activity,
        exchange=global_container.internal_exchange,
        routing_key="incoming.Activity",
    )
    mock_incoming.assert_not_awaited()


async def test_incoming_from_blocked_user_inactive_block(
    test_broker, mock_incoming, sql_session, actor_with_account
):
    remote_actor = "http://remote.test/actor"

    sql_session.add(
        Blocking(
            actor=actor_with_account,
            blocking=remote_actor,
            request="http://blocked.test/id",
            active=False,
        )
    )
    await sql_session.commit()

    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": remote_actor,
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor_with_account.actor_id,
        data=activity_pub,
    )

    await test_broker.publish(
        activity,
        exchange=global_container.internal_exchange,
        routing_key="incoming.Activity",
    )
    mock_incoming.assert_awaited_once()


async def test_incoming_message_non_gateway_actor(
    test_broker, mock_incoming, sql_session
):
    actor = await create_actor(
        sql_session, "http://localhost/", preferred_username="bob"
    )
    activity_pub = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "http://remote.test/activity_id",
        "type": "Activity",
        "actor": "http://remote.test/actor",
        "to": "http://actor.example",
        "object": "http://actor.example/object",
    }

    activity = ActivityMessage(
        actor=actor.actor_id,
        data=activity_pub,
    )
    await test_broker.publish(
        activity,
        exchange=global_container.internal_exchange,
        routing_key="incoming.Activity",
    )
    mock_incoming.assert_not_awaited()


@pytest.mark.parametrize(
    "activity, expected",
    [
        (
            {
                "type": "Activity",
            },
            True,
        ),
        (
            {
                "type": "Block",
            },
            False,
        ),
        (
            {"type": "Undo", "object": "http://follow.test/id"},
            True,
        ),
    ],
)
async def test_should_shovel_activity(sql_session, activity, expected):
    result = await should_shovel_activity(sql_session, activity)

    assert result == expected


@pytest.mark.parametrize("active, expected", [(True, False), (False, False)])
async def test_should_shovel_activity_undo(
    sql_session, actor_with_account, active, expected
):
    block_id = "http://block.test/id"
    activity = {"type": "Undo", "object": block_id}

    sql_session.add(
        Blocking(
            actor=actor_with_account,
            blocking="http://remote.test",
            request=block_id,
            active=active,
        )
    )
    await sql_session.commit()

    result = await should_shovel_activity(sql_session, activity)

    assert result == expected



from cattle_grid.extensions.examples.relationship.types import RelationshipStatus
from .status import determine_status


def test_determine_status_empty():
    result = determine_status([])

    assert result.status == RelationshipStatus.none


def test_determine_status_accepted():
    result = determine_status(
        [(False, "http://request.test/one"), (True, "http://request.test/two")]
    )

    assert result.status == RelationshipStatus.accepted
    assert result.requests == ["http://request.test/one", "http://request.test/two"]


def test_determine_status_waiting():
    result = determine_status(
        [(False, "http://request.test/one"), (False, "http://request.test/two")]
    )

    assert result.status == RelationshipStatus.waiting
    assert result.requests == ["http://request.test/one", "http://request.test/two"]



from .testing import *  # noqa


def test_get_actor_not_found(test_client):
    response = test_client.get("/html_display/html/alice")
    assert response.status_code == 404


def test_get_actor_found(test_client, publishing_actor_for_test):
    response = test_client.get(f"/html_display/html/{publishing_actor_for_test.name}")
    assert response.status_code == 200


def test_get_actor_redirect_for_activity_pub(test_client, publishing_actor_for_test):
    response = test_client.get(
        f"/html_display/html/{publishing_actor_for_test.name}",
        headers={"accept": "application/activity+json"},
        follow_redirects=False,
    )
    assert response.status_code == 307
    assert response.headers.get("location") == publishing_actor_for_test.actor
